{-# OPTIONS_GHC -fglasgow-exts -fno-warn-orphans #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Data.Trie.General.CollectionsInstances.MaybeGT
-- Copyright   :  (c) Adrian Hey 2007
-- License     :  BSD3
--
-- Maintainer  :  http://homepages.nildram.co.uk/~ahey/em.png
-- Stability   :  provisional
-- Portability :  Multi-parameter type classes, Functional dependencies
--
-- Instances of the Collections package Classes for the 'MaybeGT' type.
-----------------------------------------------------------------------------
module Data.Trie.General.CollectionsInstances.MaybeGT
(
) where

import Data.Trie.General.Types (GT(..))
import Data.Trie.General.MaybeGT

import qualified Data.Monoid as M (Monoid(..))
import qualified Data.Collections as Coll (Foldable(..),foldr',Unfoldable(..),Collection(..),Map(..))
import qualified Data.Maybe as MB (isJust)

#ifdef __GLASGOW_HASKELL__
import GHC.Base
#include "ghcdefs.h"
#else
#include "h98defs.h"
#endif

-------------------------------
-- Data.Collections.Foldable --
-------------------------------
instance GT map k => Coll.Foldable (MaybeGT map k (Maybe k,a)) (Maybe k,a) where
-- fold :: Monoid (Maybe k,a) => MaybeGT map k (Maybe k,a) -> (Maybe k,a)
   fold mp = foldrElemsAscendingMaybeGT (\assoc b -> M.mappend assoc b) mp M.mempty

-- foldMap :: Monoid m => ((Maybe k,a) -> m) -> MaybeGT map k (Maybe k,a) -> m
   foldMap f mp = foldrElemsAscendingMaybeGT (\assoc b -> M.mappend (f assoc) b) mp M.mempty

-- foldr :: ((Maybe k,a) -> b -> b) -> b -> MaybeGT map k (Maybe k,a) -> b
   foldr f b0 mp = foldrElemsAscendingMaybeGT (\assoc b -> f assoc b) mp b0

-- foldl :: (b -> (Maybe k,a) -> b) -> b -> MaybeGT map k (Maybe k,a) -> b
   foldl f b0 mp = foldrElemsDescendingMaybeGT (\assoc b -> f b assoc) mp b0

{- ToDo: Implement properly. Meantime Foldable class has suitable defaults via lists.
-- foldr1 :: ((Maybe k,a) -> (Maybe k,a) -> (Maybe k,a)) -> MaybeGT map k (Maybe k,a) -> (Maybe k,a)
-- foldl1 :: ((Maybe k,a) -> (Maybe k,a) -> (Maybe k,a)) -> MaybeGT map k (Maybe k,a) -> (Maybe k,a)
-}

-- null :: MaybeGT map k (Maybe k,a) -> Bool
   null = isEmptyMaybeGT

-- size :: MaybeGT map k (Maybe k,a) -> Int
   size mp = ASINT(addSizeMaybeGT mp L(0))

-- isSingleton :: MaybeGT map k (Maybe k,a) -> Bool
   isSingleton = isSingletonMaybeGT
-------------------------------

---------------------------------
-- Data.Collections.Unfoldable --
---------------------------------
instance GT map k => Coll.Unfoldable (MaybeGT map k (Maybe k,a)) (Maybe k,a) where
-- insert :: (Maybe k,a) -> MaybeGT map k (Maybe k,a) -> MaybeGT map k (Maybe k,a)
   insert assoc@(k,_) mp = insertMaybeGT' (const assoc) k assoc mp -- Note use of strict insertMaybeGT'

-- empty :: MaybeGT map k (Maybe k,a)
   empty = emptyMaybeGT

-- singleton :: (Maybe k,a) -> MaybeGT map k (Maybe k,a)
   singleton assoc@(k,_) = singletonMaybeGT k assoc

-- insertMany :: Foldable c' (Maybe k,a) => c' -> MaybeGT map k (Maybe k,a) -> MaybeGT map k (Maybe k,a)
   insertMany c mp0 = Coll.foldr (\assoc@(k,_) mp -> insertMaybeGT' (const assoc) k assoc mp) mp0 c -- ?? stricness?? l/r??

-- insertManySorted :: Foldable c' (Maybe k,a) => c' -> MaybeGT map k (Maybe k,a) -> MaybeGT map k (Maybe k,a)
   insertManySorted c mp0 = Coll.foldr (\assoc@(k,_) mp -> insertMaybeGT' (const assoc) k assoc mp) mp0 c -- How to implement efficiently ??
---------------------------------


---------------------------------
-- Data.Collections.Collection --
---------------------------------
instance GT map k => Coll.Collection (MaybeGT map k (Maybe k,a)) (Maybe k,a) where
-- filter :: ((Maybe k,a) -> Bool) -> MaybeGT map k (Maybe k,a) -> MaybeGT map k (Maybe k,a)
   filter = filterMaybeGT
---------------------------------


--------------------------
-- Data.Collections.Map --
--------------------------
instance (GT map k, M.Monoid a) => Coll.Map (MaybeGT map k a) (Maybe k) a where
-- delete :: Maybe k -> MaybeGT map k a -> MaybeGT map k a
   delete = deleteMaybeGT

-- member :: Maybe k -> MaybeGT map k a -> Bool
   member k mp = MB.isJust (lookupMaybeGT k mp)

-- union :: MaybeGT map k a -> MaybeGT map k a -> MaybeGT map k a
   union = unionMaybeGT' (\x _ -> x) -- Note use of strict unionMaybeGT'

-- intersection :: MaybeGT map k a -> MaybeGT map k a -> MaybeGT map k a
   intersection = intersectionMaybeGT' (\x _ -> x)  -- Note use of strict intersectionMaybeGT'

-- difference :: MaybeGT map k a -> MaybeGT map k a -> MaybeGT map k a
   difference = differenceMaybeGT

-- isSubset :: MaybeGT map k a -> MaybeGT map k a -> Bool
   isSubset = isSubsetOfMaybeGT

-- lookup :: Monad m => Maybe k -> MaybeGT map k a -> m a
   lookup k mp = case lookupMaybeGT k mp of
                 Just a  -> return a
                 Nothing -> fail "Data.Collections.Map.lookup: Key not found in MaybeGT."

-- alter :: (Maybe a -> Maybe a) -> Maybe k -> MaybeGT map k a -> MaybeGT map k a
   alter = alterMaybeGT

-- insertWith :: (a -> a -> a) -> Maybe k -> a -> MaybeGT map k a -> MaybeGT map k a
   insertWith f k a ogt = insertMaybeGT (f a) k a ogt

-- fromFoldableWith :: Foldable l (Maybe k,a) => (a -> a -> a) -> l -> MaybeGT map k a
   fromFoldableWith f l = Coll.foldr insrt emptyMaybeGT l  -- Strictness ??
    where insrt (k,a) ogt = insertMaybeGT (f a) k a ogt

-- foldGroups :: Foldable l (Maybe k,b) => (b -> a -> a) -> a -> l -> MaybeGT map k a
   foldGroups f a0 l = Coll.foldr' insrt emptyMaybeGT l
    where insrt (k,b) ogt = insertMaybeGT (f b) k (f b a0) ogt

-- mapWithKey :: (Maybe k -> a -> a) -> MaybeGT map k a -> MaybeGT map k a
   mapWithKey = mapWithKeyMaybeGT

-- unionWith :: (a -> a -> a) -> MaybeGT map k a -> MaybeGT map k a -> MaybeGT map k a
   unionWith = unionMaybeGT

-- intersectionWith :: (a -> a -> a) -> MaybeGT map k a -> MaybeGT map k a -> MaybeGT map k a
   intersectionWith = intersectionMaybeGT

-- differenceWith :: (a -> a -> Maybe a) -> MaybeGT map k a -> MaybeGT map k a -> MaybeGT map k a
   differenceWith = differenceMaybeMaybeGT

-- isSubmapBy :: (a -> a -> Bool) -> MaybeGT map k a -> MaybeGT map k a -> Bool
   isSubmapBy = isSubmapOfMaybeGT
--------------------------
